import os
import json
from pybythec import utils
import commentjson as json

log = utils.log


class BuildStatus:
    '''
        member variables:
        name: name of target
        path: path to write status.json file to
        status: failed, built, up to date, or locked
        description: what happened
    '''
    def __init__(self, name, path = ''):
        self.name = name
        self.path = path
        self.status = 'failed'
        self.description = ''

    def readFromFile(self, libSrcDir, buildDir, binaryRelfPath):
        '''
            buildPath (in): where to read the status.json file from
        '''

        libSrcDirShellPath = utils.getShellPath(libSrcDir)

        buildPath = f'{libSrcDirShellPath}/{buildDir}/{binaryRelfPath}'

        if not os.path.exists(buildPath):
            # try the other hidden / non-hidden version
            if buildDir[0] == '.':
                buildPath = f'{libSrcDirShellPath}/{buildDir.lstrip(".")}/{binaryRelfPath}'
            else:
                buildPath = f'{libSrcDirShellPath}/.{buildDir}/{binaryRelfPath}'

        with open(buildPath + '/status.json', 'r') as rf:
            contents = json.load(rf)
        # contents = utils.loadJsonFile(buildPath + '/status.json')
        
        if not contents:
            self.description = f'couldn\'t find contents in {buildPath}'
            log.error(f'couldn\'t find contents in {buildPath}')
            return
        if 'status' in contents:
            self.status = contents['status']
        else:
            self.description = f'couldn\'t find the build status in {buildPath}'
            log.error(self.description)
        if 'description' in contents:
            self.description = contents['description']
        else:
            self.description = f'{buildPath} doesn\'t contain a description'
            log.warning(self.description)

    def writeInfo(self, status, msg):
        log.info(msg)
        self.status = status
        self.description = msg
        self._writeToFile()

    def writeError(self, msg):
        log.error(msg)
        self.description = msg
        self._writeToFile()

    def _writeToFile(self):
        if not os.path.exists(self.path):
            return
        with open(self.path + '/status.json', 'w') as f:
            json.dump({'status': self.status, 'description': self.description}, f, indent = 4)



# -*- coding: utf-8 -*-
from pybythec import utils
from pybythec.utils import PybythecError
from pybythec.BuildStatus import BuildStatus
from pybythec.BuildElements import BuildElements
from multiprocessing import Process, Queue
from threading import Thread
import traceback
import os
import time

log = utils.log

__author__ = 'glowtree'
__email__ = 'tom@glowtree.com'
__version__ = '0.9.64'


def getBuildElements(osType = None,
                     compiler = None,
                     buildType = None,
                     binaryFormat = None,
                     projConfigPath = None,
                     globalConfigPath = None,
                     projConfig = None,
                     globalConfig = None,
                     currentBuild = None,
                     libDir = None):
    '''
        passthrough function that catches and reports exceptions
    '''
    try:
        return BuildElements(osType = osType,
                             compiler = compiler,
                             buildType = buildType,
                             binaryFormat = binaryFormat,
                             projConfig = projConfig,
                             projConfigPath = projConfigPath,
                             globalConfig = globalConfig,
                             globalConfigPath = globalConfigPath,
                             currentBuild = currentBuild,
                             libDir = libDir)
    except PybythecError as e:
        log.error(e)
        return None
    except Exception as e:
        log.error(f'unknown exception: {str(e)}\n{traceback.format_exc()}')
        return None


def build(be = None, builds = None):
    '''
        be: BuildElements object
        builds: list of build overrides
    '''
    if not be:
        be = getBuildElements()
        if not be:
            return

    _runPreScript(be)

    buildsRef = builds
    if not buildsRef:
        buildsRef = be.builds
    if type(buildsRef) is not list:
        buildsRef = [buildsRef]

    for br in buildsRef:
        try:
            be.configBuild(currentBuild = br)
        except PybythecError as e:
            log.error(e)
            continue
        except Exception as e:
            log.error(f'unknown exception: {str(e)}\n{traceback.format_exc()}')
            continue
        log.debug(br)
        _build(be)


def _build(be):
    '''
        does the dirty work of compiling and linking based on the state setup in the BuildElements object be
    '''

    buildStatus = BuildStatus(be.targetFilename, be.buildDirPathShell)

    # lock - early return
    if be.locked and utils.pathExists(be.installPathShell):
        buildStatus.writeInfo('locked', f'{be.targetName} is locked')
        return True

    startTime = time.time()

    log.info(f'building {be.infoStr}')

    buildingLib = False
    if be.libDir:
        buildingLib = True

    if not utils.pathExists(be.buildDirPathShell):
        utils.createDirs(be.buildDirPathShell)

    if not utils.pathExists(be.installDirPathShell):
        utils.createDirs(be.installDirPathShell)

    incPathList = []
    for incPath in be.incPaths:
        if utils.pathExists(incPath):
            incPathList += ['-I', incPath]
        else:
            log.warning(f'incPath {incPath} doesn\'t exist')

    for extIncPath in be.extIncPaths:  # external include libs (for cases where 3rd party header includes are using "" instead of <> ie Unreal)
        if utils.pathExists(extIncPath):
            incPathList += ['-I', extIncPath]
        else:
            log.warning(f'extIncPath {extIncPath} doesn\'t exist')

    definesList = []
    for define in be.defines:
        definesList += ['-D', define]

    #
    # qt moc file compilation, TODO: make this another compiler option, along with asm
    #
    mocPaths = []
    for qtClass in be.qtClasses:
        found = False
        mocPath = f'{be.buildDirPath}/moc_{qtClass}.cpp'
        qtClassHeader = qtClass + '.h'

        for incPath in be.incPaths:  # find the header file, # TODO: should there be a separate list of headers ie be.mocIncPaths?
            includePath = f'{incPath}/{qtClassHeader}'
            if not utils.pathExists(includePath):
                continue

            if utils.pathExists(mocPath) and float(os.stat(mocPath).st_mtime) < float(os.stat(includePath).st_mtime) or not utils.pathExists(mocPath):
                buildStatus.description = 'qt moc: ' + utils.runCmd(['moc'] + definesList + [includePath, '-o', mocPath])

            if not utils.pathExists(mocPath):
                buildStatus.writeError(buildStatus.description)
                return False

            mocPaths.append(mocPath)
            found = True

        if not found:
            buildStatus.writeError(f'can\'t find {qtClassHeader} for qt moc compilation')
            return False

    for mocPath in mocPaths:
        be.sources.append(mocPath)

    threads = []
    buildStatusQueue = Queue()  # the build status for each dependency: objs and libs
    objPathQueue = Queue()

    if be.processType is utils.MULTI_PROCESS:
        log.debug(f'using multiprocessing')
    elif be.processType is utils.MULTI_THREAD:
        log.debug(f'using multithreading')
    else:
        log.debug(f'using single process')
    
    #
    # compile
    #
    objPaths = []
    cmd = [be.compilerCmd, be.objFlag] + incPathList + definesList + be.flags

    for source in be.sources:
        if be.processType is utils.MULTI_PROCESS:
            thread = Process(target = _compileSrc, args = (be, cmd, source, objPathQueue, buildStatusQueue))
        elif be.processType is utils.MULTI_THREAD:
            thread = Thread(target = _compileSrc, args = (be, cmd, source, objPathQueue, buildStatusQueue))
        else:
            _compileSrc(be, cmd, source, objPathQueue, buildStatusQueue)
        if be.processType > 0:
            thread.start()
            threads.append(thread)

    #
    # build library dependencies
    #
    libCmds = []
    if be.binaryType == 'exe' or be.binaryType == 'plugin':
        for lib in be.libs:
            libName = lib
            if be.compiler.startswith('msvc'):
                libCmds += [libName + be.staticExt]  # you need to link against the .lib stub file even if it's ultimately a .dll that gets linked
            else:
                libCmds += [be.libFlag, libName]

    # check if the lib has a directory for building
    for libSrcPath in be.libSrcPaths:
        if utils.pathExists(libSrcPath):
            if be.processType is utils.MULTI_PROCESS:
                thread = Process(target = _buildLib, args = (be, libSrcPath, buildStatusQueue))
            elif be.processType is utils.MULTI_THREAD:
                thread = Thread(target = _buildLib, args = (be, libSrcPath, buildStatusQueue))
            else:
                _buildLib(be, libSrcPath, buildStatusQueue)
            if be.processType > 0:
                thread.start()
                threads.append(thread)

    # wait for all the threads before checking the results
    for thread in threads:
        thread.join()

    allUpToDate = True
    while not buildStatusQueue.empty():
        bs = buildStatusQueue.get()
        if bs.status == 'failed':
            # NOTE: changed from bs.description.encode('ascii', 'ignore') which fixed issue on macOs
            totalTime = int(time.time() - startTime)
            buildStatus.writeError(
                f'{be.infoStr} failed because {bs.name} failed because...\n\n{bs.description}\n...determined in {totalTime} seconds\n\n')
            return False
        elif bs.status == 'built':
            allUpToDate = False

    # objPaths = []
    while not objPathQueue.empty():
        objPaths.append(objPathQueue.get())

    # revise the library paths
    extraLibPaths = []
    for i in range(len(be.libPaths)):

        revisedLibPath = f'{be.libPaths[i]}/{be.osType}/{be.compilerVersion}'
        if utils.pathExists(revisedLibPath):
            be.libPaths[i] = revisedLibPath

        revisedLibPath_ = f'{revisedLibPath}/{be.binaryFormat}'
        if utils.pathExists(revisedLibPath_):
            revisedLibPath = revisedLibPath_
            extraLibPaths.append(revisedLibPath)

        revisedLibPath_ = f'{revisedLibPath}/{be.buildType}'
        if utils.pathExists(revisedLibPath_):
            revisedLibPath = revisedLibPath_
            extraLibPaths.append(revisedLibPath)

        if be.currentBuild:
            revisedLibPath += f'/{be.currentBuild}'
            if utils.pathExists(revisedLibPath):
                extraLibPaths.append(revisedLibPath)

    for xlp in extraLibPaths:
        be.libPaths.insert(0, xlp)

    #
    # linking
    #
    linkCmd = []

    if allUpToDate and utils.pathExists(be.installPathShell):
        totalTime = int(time.time() - startTime)
        buildStatus.writeInfo('up to date', f'{be.infoStr} is up to date, determined in {totalTime} seconds\n')
        if not buildingLib:
            _runPostScript(be)
        return True

    # microsoft's compiler / linker can only handle so many characters on the command line
    if be.compiler.startswith('msvc'):
        objPathsStr = '" "'.join(objPaths)
        libCmdsStr = ' '.join(libCmds)
        msvcLinkCmd = f'{be.targetFlag}"{be.installPath}" "{objPathsStr}" {libCmdsStr}'
        with open(be.buildDirPathShell + '/linkCmd', 'w') as wf:
            wf.write(msvcLinkCmd)

        linkCmd += [be.linker, '@' + be.buildDirPath + '/linkCmd']
        if be.showLinkerCmds:
            log.info(f'\nmsvcLinkCmd: {msvcLinkCmd}\n')
    else:
        linkCmd += [be.linker, be.targetFlag, be.installPath] + objPaths + libCmds

    if be.binaryType != 'static':
        linkCmd += be.linkFlags

    if be.binaryType == 'exe' or be.binaryType == 'plugin' or (be.compilerRoot == 'msvc' and be.binaryType == 'dynamic'):

        for libPath in be.libPaths:
            if not utils.pathExists(libPath):
                log.warning(f'libPath {libPath} doesn\'t exist')
                continue
            if be.compiler.startswith('msvc'):
                linkCmd += [be.libPathFlag + os.path.normpath(libPath)]
            else:
                linkCmd += [be.libPathFlag, os.path.normpath(libPath)]

    # get the timestamp of the existing target if it exists
    linked = False
    targetExisted = False
    oldTargetTimeStamp = None
    if utils.pathExists(be.installPathShell):
        oldTargetTimeStamp = float(os.stat(be.installPathShell).st_mtime)
        targetExisted = True

    if be.showLinkerCmds:
        linkCmdStr = ' '.join(linkCmd)
        log.info(f'\n{linkCmdStr}\n')

    buildStatus.description = utils.runCmd(linkCmd)

    if utils.pathExists(be.installPathShell):
        if targetExisted:
            if float(os.stat(be.installPathShell).st_mtime) > oldTargetTimeStamp:
                linked = True
        else:
            linked = True

    if linked:
        log.info(f'linked {be.infoStr}')
    else:
        buildStatus.writeError(f'linking failed because {buildStatus.description}')
        return False

    # copy dynamic library dependencies to the install path
    if be.copyDynamicLibs:
        if be.binaryType == 'exe' or be.binaryType == 'plugin':
            for lib in be.libs:
                for libPath in be.libPaths:
                    dynamicPath = libPath + '/'
                    if be.compilerRoot == 'gcc' or be.compilerRoot == 'clang':
                        dynamicPath += 'lib'
                    dynamicPath += lib + be.dynamicExt
                    dynamicPath = utils.getShellPath(dynamicPath)
                    if utils.pathExists(dynamicPath):
                        utils.copyfile(dynamicPath, be.installDirPathShell)

    totalTime = int(time.time() - startTime)
    buildStatus.writeInfo('built', f'{be.infoStr} built {be.installPath}\ncompleted in {totalTime} seconds\n')

    # run a post-build script if it exists
    if not buildingLib:
        _runPostScript(be)

    return True


#
# private functions
#
def _compileSrc(be, compileCmd, source, objPathQueue, buildStatusQueue):
    '''
        be (in): BuildElements object
        compileCmd (in): the compile command so far
        source (in): the c or cpp source file to compile (every source file gets it's own object file)
        objPaths (out): list of all object paths that will be passed to the linker
        buildStatus (out): build status for this particular compile, defaults to failed
    '''

    buildStatus = BuildStatus(source)

    if not utils.pathExists(source):
        buildStatus.writeError(f'{source} is missing, exiting build')
        buildStatusQueue.put(buildStatus)
        return

    objFile = os.path.basename(source)
    objFile = objFile.replace(os.path.splitext(source)[1], be.objExt)
    objPath = f'{be.buildDirPath}/{objFile}'
    objPathQueue.put(objPath)

    # check if it's up to date
    objExisted = utils.pathExists(objPath)
    if objExisted:
        objTimestamp = float(os.stat(utils.getShellPath(objPath)).st_mtime)
        if objTimestamp > be.latestConfigTimestamp and not utils.sourceNeedsBuilding(be.incPaths, source, objTimestamp):
            buildStatus.status = 'up to date'
            buildStatusQueue.put(buildStatus)
            return

        # if not utils.sourceNeedsBuilding(be.incPaths, source, objTimestamp):
        #   buildStatus.status = 'up to date'
        #   return

    # Microsoft Visual C has to have the objPathFlag cuddled up directly next to the objPath - no space in between them (grrr)
    if be.compiler.startswith('msvc'):
        cmd = compileCmd + [source, be.objPathFlag + objPath]
    else:
        cmd = compileCmd + [source, be.objPathFlag, objPath]

    if be.showCompilerCmds:
        cmdStr = ' '.join(cmd)
        log.info(f'\n{cmdStr}\n')

    buildStatus.description = utils.runCmd(cmd)

    if utils.pathExists(objPath):
        if objExisted:
            if float(os.stat(utils.getShellPath(objPath)).st_mtime) > objTimestamp:
                buildStatus.status = 'built'
        else:
            buildStatus.status = 'built'

    if buildStatus.status == 'built':
        buildStatus.description = 'compiled ' + os.path.basename(source)
    else:
        log.error(f'{objPath} failed to build')

    buildStatusQueue.put(buildStatus)


def _buildLib(be, libSrcDir, buildStatusQueue):
    '''
    '''
    libBe = getBuildElements(osType = be.osType,
                             compiler = be.compiler,
                             buildType = be.buildType,
                             binaryFormat = be.binaryFormat,
                             projConfig = be.projConfig,
                             globalConfig = be.globalConfig,
                             currentBuild = be.currentBuild,
                             libDir = libSrcDir)
    if not libBe:
        return

    build(libBe)

    buildStatus = BuildStatus(libSrcDir)
    buildStatus.readFromFile(libSrcDir, be.buildDir, be.binaryRelPath)
    buildStatusQueue.put(buildStatus)


def clean(be = None, builds = None):
    '''
    '''
    if not be:
        be = getBuildElements()
        if not be:
            return

    buildsRef = builds
    if not buildsRef:
        buildsRef = be.builds
    if type(buildsRef) is not list:
        buildsRef = [buildsRef]

    for build in buildsRef:
        try:
            be.configBuild(currentBuild = build)
        except PybythecError as e:
            log.error(e)
            return
        except Exception as e:
            log.error(f'unknown exception: {str(e)}\n{traceback.format_exc()}')
            return
        _clean(be)


def _clean(be = None):
    '''
        cleans the current project
        be (in): BuildElements object
    '''

    # remove any dynamic libs that are sitting next to the exe
    if utils.pathExists(be.installDirPathShell) and (be.binaryType == 'exe' or be.binaryType == 'plugin'):
        for fl in os.listdir(be.installDirPathShell):
            libName, ext = os.path.splitext(fl)
            if ext == be.dynamicExt:
                if be.compilerRoot == 'gcc' or be.compilerRoot == 'clang':
                    libName = libName.lstrip('lib')
                for lib in be.libs:
                    if lib == libName:
                        p = be.installDirPathShell + '/' + fl
                        try:
                            os.remove(p)
                        except Exception:
                            log.warning(f'failed to remove {p}')
            elif ext == '.exp' or ext == '.ilk' or ext == '.lib' or ext == '.pdb':  # msvc files
                p = be.installDirPathShell + '/' + fl
                try:
                    os.remove(p)
                except Exception:
                    log.warning(f'failed to remove {p}')

    if not utils.pathExists(be.buildDirPathShell):  # canary in the coal mine
        log.info(f'{be.infoStr} already clean')
        return True

    dirCleared = True
    for fl in os.listdir(be.buildDirPathShell):
        p = be.buildDirPathShell + '/' + fl
        try:
            os.remove(p)
        except Exception:
            dirCleared = False
            log.warning(f'failed to remove {p}')
    if dirCleared:
        os.removedirs(be.buildDirPathShell)

    if utils.pathExists(be.installPathShell):
        os.remove(be.installPathShell)
    target, ext = os.path.splitext(be.installPathShell)
    if ext == '.dll':
        try:
            os.remove(target + '.exp')
            os.remove(target + '.lib')
        except Exception:
            pass
    try:
        os.removedirs(be.installDirPathShell)
    except Exception:
        pass

    log.info(f'{be.infoStr} all clean')
    return True


def cleanAll(be = None, builds = None):
    '''
        cleans both the current project and also the dependencies
    '''
    if not be:
        be = getBuildElements()
        if not be:
            return

    buildsRef = builds
    if not buildsRef:
        buildsRef = be.builds
    if type(buildsRef) is not list:
        buildsRef = [buildsRef]

    for build in buildsRef:
        try:
            be.configBuild(currentBuild = build)
        except PybythecError as e:
            log.error(e)
            continue
        except Exception as e:
            log.error(f'unknown exception: {str(e)}\n{traceback.format_exc()}')
            continue
        _clean(be)

        # clean library dependencies
        for libSrcPath in be.libSrcPaths:
            if utils.pathExists(libSrcPath):
                libBe = getBuildElements(osType = be.osType,
                                            compiler = be.compiler,
                                            buildType = be.buildType,
                                            binaryFormat = be.binaryFormat,
                                            projConfig = be.projConfig,
                                            globalConfig = be.globalConfig,
                                            currentBuild = be.currentBuild,
                                            libDir = libSrcPath)
                if not libBe:
                    return
                clean(libBe)  # builds = build)


def _runPreScript(be):
    '''
        looks for a pre-build script and loads it as a module
    '''
    pathRoot = '.'
    if be.libDir:
        pathRoot = be.libDir
    preScriptPath = utils.getShellPath(pathRoot + '/pybythecPre.py')
    if not utils.pathExists(preScriptPath):
        preScriptPath = pathRoot + '/.pybythecPre.py'
    if utils.pathExists(preScriptPath):
        import imp
        m = imp.load_source('', preScriptPath)
        m.run(be)


def _runPostScript(be):
    '''
        looks for a post-build script and loads it as a module
    '''
    pathRoot = '.'
    if be.libDir:
        pathRoot = be.libDir
    postScriptPath = utils.getShellPath(pathRoot + '/pybythecPost.py')
    if not utils.pathExists(postScriptPath):
        postScriptPath = pathRoot + '/.pybythecPost.py'
    if utils.pathExists(postScriptPath):
        import importlib.util
        spec = importlib.util.spec_from_file_location('', postScriptPath)
        if not spec:
            return
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        module.run(be)

        # m = importlib.load_source('', postScriptPath)
        # import imp
        # m = imp.load_source('', postScriptPath)
        # m.run(be)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
test_pybythec
----------------------------------

tests for pybythec module
'''

import os
import unittest
import subprocess
import pybythec

log = pybythec.utils.log


class TestPybythec(unittest.TestCase):
    def setUp(self):
        '''
            typical setup for building with pybythec
        '''

        # setup the environment variables...
        # normally you would probably set these in your .bashrc (linux / macOs), profile.ps1 (windows) file etc
        cwd = os.getcwd()
        cwd = cwd.replace('\\', '/')
        os.environ['PYBYTHEC_EXAMPLE_SHARED'] = cwd + '/example/shared'
        os.environ['PYBYTHEC_GLOBALS'] = cwd + '/globals.json'  # this overrides ~/.pybythecGlobals

    def test_000_something(self):
        '''
            build
        '''
        # print('\n')
        log.raw('\n')
        log.info(f'cwd: {os.getcwd()}')

        # build Plugin
        log.raw('\n')
        os.chdir('./example/projects/Plugin')
        pybythec.build()
        log.raw('\n')

        # build Main (along with it's library dependencies)
        log.raw('\n')
        os.chdir('../Main')
        be = pybythec.getBuildElements()
        pybythec.build(be)
        log.raw('\n')

        for b in be.builds:

            # log.debug(b)
            # continue

            exePath = f'./{b}/Main'
            if be.osType == 'windows':
                exePath += '.exe'

            log.info(f'checking that {exePath} exists...')
            self.assertTrue(os.path.exists(exePath))

            p = subprocess.Popen([exePath], stdout = subprocess.PIPE, stderr = subprocess.PIPE)
            stdout, stderr = p.communicate()
            stdout = stdout.decode('utf-8')
            log.info(stdout)

            if len(stderr):
                raise Exception(stderr)

            self.assertTrue(
                stdout.startswith('running an executable and a statically linked library and a dynamically linked library'))  # and a plugin'))
            
            log.raw('\n')

    def tearDown(self):
        '''
            clean the builds
        '''
        pybythec.cleanAll()

        os.chdir('../Plugin')
        pybythec.cleanAll()


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



# import json
import shutil
import subprocess
import platform
import sys
import os

LINUX_ROOT = ''
if 'microsoft-standard' in platform.uname(
).release:  # check if it's windows subsystem for linux
    LINUX_ROOT = '/mnt'

LINEAR_PROCESS = 0
MULTI_THREAD = 1
MULTI_PROCESS = 2


class PybythecError(Exception):
    def __init__(self, msg):
        super(PybythecError, self).__init__(msg)


class Logger:
    def __init__(self, name = None, debug = False, filepath = None):
        self.name = name
        if self.name:
            self.name += ': '
        else:
            self.name = ''
        self._debug = debug
        self.wf = sys.stdout
        if filepath:
            self.toFile = True
            self.wf = open(filepath, 'w')
        else:
            self.toFile = False

    def __del__(self):
        if self.toFile:
            self.wf.close()

    def debug(self, s):
        if self._debug:
            self.wf.write(f'debug: {self.name}{s}\n')
            self.wf.flush()

    def info(self, s):
        self.wf.write(f'{self.name}{s}\n')
        self.wf.flush()

    def warning(self, s):
        self.wf.write(f'warning: {self.name}{s}\n')
        self.wf.flush()

    def error(self, s):
        if self.toFile:
            self.wf.write(f'error: {self.name}{s}\n')
            self.wf.flush()
        else:
            sys.stderr.write(f'error: {self.name}{s}\n')
            sys.stderr.flush()

    def raw(self, s):
        sys.stdout.write(s)

    # shorthands
    def d(self, s):
        self.debug(s)

    def i(self, s):
        self.info(s)

    def w(self, s):
        self.warning(s)

    def e(self, s):
        self.error(s)

    def r(self, s):
        self.raw(s)


log = Logger()  # singleton


def srcNewer(srcPath, dstPath):
    if int(os.stat(srcPath).st_mtime) > int(os.stat(dstPath).st_mtime):
        return True
    return False


def checkTimestamps(incPaths, src, timestamp):
    '''
    finds the newest timestamp of everything upstream of the src file, including the src file
  '''
    srcPath = getShellPath(src)
    if not os.path.exists(srcPath):
        log.warning(f'checkTimestamps: {srcPath} doesn\'t exist')
        return

    srcTimeStamp = float(os.stat(srcPath).st_mtime)
    if srcTimeStamp > timestamp[0]:
        timestamp[0] = srcTimeStamp

    fileCopy = str()
    srcFile = open(srcPath, 'r')
    for line in srcFile:
        fileCopy += line
    srcFile.close()

    for line in fileCopy.split('\n'):
        if line.startswith('#include'):
            filename = line.lstrip('#include')
            filename = filename.strip()
            if (filename[0] == '"'):
                filename = filename.strip('"')
                for dir in incPaths:
                    filepath = f'{dir}/{filename}'
                    if os.path.exists(filepath):
                        checkTimestamps(incPaths, filepath, timestamp)


def sourceNeedsBuilding(incPaths, src, objTimestamp):
    '''
    determines whether a source file needs to be built or not
  '''
    timestamp = [0]  # [] so it's passed as a reference
    checkTimestamps(incPaths, src, timestamp)

    if timestamp[0] > objTimestamp:
        return True

    return False


def getLibPath(libName, libPath, compiler, libExt):
    '''
      get the lib path with the os / compiler specific prefix and file extension
  '''
    libPath += '/'
    if compiler.startswith('gcc') or compiler.startswith('clang'):
        libPath += 'lib'
    libPath += libName + libExt
    return libPath


def isWindowsPath(path):
    '''
    path: assume it's an absolute path ie C:/hi
  '''
    if len(path) < 3:
        return False
    if path[0].isalpha() and path[1] == ':' and (path[2] == '/'
                                                 or path[2] == '\\'):
        return True
    return False


def windowsToLinux(p):
    '''
  '''
    np = p.replace('\\', '/')
    driveLetter = np[0].lower()
    return f'{LINUX_ROOT}/{driveLetter}{np[2:]}'


def linuxToWindows(p):
    '''
  '''
    np = p[len(LINUX_ROOT) + 1:]
    return np[0].upper() + ':' + np[1:]


def getShellPath(path):
    '''
    path: assume absolute path
    returns the usable path for the current shell
  '''
    if not isWindowsPath(path):
        return path
    if platform.system() == 'Linux' or platform.system() == 'Darwin':
        return windowsToLinux(path)
    return path


def pathExists(path):
    '''
  '''
    return os.path.exists(getShellPath(path))


def getPath(path, osType):
    '''
    gets the path based on the requested os type ie windows, linux
  '''
    isWin = isWindowsPath(path)
    if osType == 'windows':
        if isWin:
            return path
        return linuxToWindows(path)
    # otherwise linux
    if isWin:
        return windowsToLinux(path)
    return path


def getShellOsType():
    '''
  '''
    if platform.system() == 'Linux':
        return 'linux'
    elif platform.system() == 'Darwin':
        return 'macOs'
    elif platform.system() == 'Windows':
        return 'windows'
    else:
        raise PybythecError('os needs to be linux, macOs or windows')


def _getAbsPath(cwDir, path):
    '''
  '''
    _cwDir = cwDir.replace('\\', '/')
    if not _cwDir.endswith('/'):
        _cwDir += '/'
    _path = path.replace('\\', '/')
    if len(path) < 2:
        if path[0] == '.':
            return _cwDir.rstrip('/')
    if _path[0] == '.' and path[1].isalpha():  # ie .pybythec
        return _cwDir + _path
    return os.path.normpath(os.path.join(cwDir, './' + path)).replace('\\', '/')
    # TODO: handle the case when it's ./.pybythec
    # return _cwDir + _path.lstrip('./\\')


def getAbsPath(cwDir, path):
    '''
    cwDir: current working dir path
    path: may be relative or absolute
    returns absolute path
  '''
    if isWindowsPath(path):
        return path.replace('\\', '/')
    if os.path.isabs(path):
        return path
    return _getAbsPath(cwDir, path)


def resolvePaths(cwDir, paths, osType):
    '''
    '''
    i = 0
    for path in paths:
        p = getAbsPath(cwDir, path)
        paths[i] = getPath(p, osType)
        i += 1


def createDirs(path):
    '''
        recursively goes up the path heiarchy creating the necessary directories along the way
        similar to os.makedirs except doesn't throw an exception if a directory's already exists
        also os.makedirs throws the same exception whether the directory already exists or it couldn't create it, not ideal
    '''
    if path is None or not len(path):
        log.warning('createDirs: empty path')
        return

    # in case path ends with a '/'
    path = path.rstrip('/')

    if os.path.exists(path):
        return

    # if the path above the current one doesn't exist, create it
    abovePath = os.path.dirname(path)
    if not os.path.exists(abovePath):
        createDirs(abovePath)

    try:
        os.mkdir(path)
    except Exception:  # OSError:
        # log.warning('failed to make {0} because {1}', path, e)
        pass


def copyfile(srcPath, dstDir):
    '''
        copies srcPath to dstPath, creating the directory structure if necessary for the destination
        srcPath: absolute file path
        dstDir:  absolute directory path
    '''

    if not os.path.exists(srcPath):
        return False

    dstPath = f'{dstDir}/{os.path.basename(srcPath)}'

    if os.path.exists(dstPath):
        if not srcNewer(srcPath, dstPath):
            return

    # in case the path doesn't already exist
    createDirs(dstDir)

    shutil.copy2(srcPath, dstDir)

    log.debug(f'{srcPath} copied to {dstPath}')

    return True


# def loadJsonFile(jsonPath):
#     '''
#     load a json config file
#     NOTE: no check for existence of the path so that logging warnings can be controlled elsewhere
#   '''
#     if os.path.splitext(jsonPath)[1] != '.json':
#         # raise PybythecError(f'{jsonPath} is not a json file')
#         return None
#     if not os.path.exists(jsonPath):
#         raise PybythecError(f'{jsonPath} doesn\'t exist')
#     try:
#         with open(jsonPath) as f:
#             return json.loads(removeComments(f))
#     except Exception as e:
#         raise PybythecError(f'failed to parse {jsonPath}: {e}')

# def removeComments(f):
#     '''
#     removes // style comments from a file, num of lines stays the same
#   '''
#     sansComments = ''
#     inQuotes = False
#     for l in f:
#         i = 0
#         for c in l:
#             if c == '"':
#                 inQuotes = not inQuotes
#             elif c == '/' and l[i + 1] == '/' and not inQuotes:
#                 sansComments += '\n'
#                 break
#             i += 1
#             sansComments += c
# return sansComments


def runCmd(cmd):
    '''
    runs a command and blocks until it's done, returns the output
  '''
    try:
        p = subprocess.Popen(cmd,
                             stdout = subprocess.PIPE,
                             stderr = subprocess.PIPE)
    except subprocess.CalledProcessError as e:
        return f'cmd failed: {" ".join(cmd)} because: {e.output}'
    except Exception:
        return f'cmd failed: {" ".join(cmd)}'
    stdout, stderr = p.communicate()
    output = ''
    if len(stderr):
        output += stderr.decode('utf-8')
    if len(stdout):
        output += stdout.decode('utf-8')
    return output


# testing
if __name__ == '__main__':

    print(
        getAbsPath(
            'C:\\Users\\tom\\work_offline\\repos\\pybythec/example/shared/src\\DynamicLib',
            '../../include'))



DEBUG = False

import os
import subprocess
from pybythec import utils
from pybythec.utils import PybythecError
import commentjson as json

log = utils.log


class BuildElements:
    def __init__(self,
                 osType = None,
                 compiler = None,
                 buildType = None,
                 binaryFormat = None,
                 projConfigPath = None,
                 globalConfigPath = None,
                 projConfig = None,
                 globalConfig = None,
                 currentBuild = None,
                 libDir = None):
        '''
            osType: operating system: currently linux, macOs, or windows
            compiler: any variation of gcc, clang, or msvc ie gcc-4.4, msvc110, if plusplus is true gcc, clang will become g++, clang++
            buildType: debug release etc
            binaryFormat: 32bit, 64bit etc
            projConfigPath: path to a pybythec project config file (json)
            globalConfigPath: path to a pybythec global config file (json)
            projConfig: dict of the project config
            globalConfig: dict of the global config
            libDir: directory path of the library being built, likely only used when building a library as a dependency (ie from a project)
            currentBuild: current build from a potential list of custom builds, passed in when building library dependencies
            parses config files: global, project, local if they exist to determine build elements (state), uses function arguments as overrides
        '''

        self.localConfig = None
        self.projConfig = projConfig
        self.globalConfig = globalConfig
        self.currentBuild = currentBuild  # current custom build if any
        self.libDir = libDir

        # overrides config files
        self.osTypeOverride = osType
        self.compilerOverride = compiler
        self.buildTypeOverride = buildType
        self.binaryFormatOverride = binaryFormat

        self.buildType = None
        self.version = None
        self.targetName = None  # name of the target
        self.targetFilename = None  # name of the target + extension
        self.builds = None  # a list of custom build keys

        self.osType = None  # linux, macOs, windows
        self.binaryType = None  # exe, static, dynamic, plugin

        self.version = 0
        self.binaryFormat = None  # 32bit, 64bit etc
        self.binaryFormatDefault = '64bit'
        self.libInstallPathAppend = True
        self.plusplus = True
        self.locked = False
        self.buildDir = '.'
        self.showCompilerCmds = False
        self.showLinkerCmds = False

        self.copyDynamicLibs = False

        self.msvcDefault = None

        self.cwDir = None
        self.shellCwDir = os.getcwd()
        if self.libDir:
            self.shellCwDir = utils.getShellPath(self.libDir)

        self.buildDirPath = None
        self.buildDirPathShell = None

        self.installDirPath = None
        self.installDirPathShell = None

        self.installPath = None
        self.installPathShell = None

        self.latestConfigTimestamp = 0

        self.defines = []

        self.processType = utils.MULTI_THREAD #  LINEAR_PROCESS MULTI_THREAD MULTI_PROCESS 

        # global config
        if not self.globalConfig:
            if 'PYBYTHEC_GLOBALS' in os.environ:
                globalConfigPath = os.environ['PYBYTHEC_GLOBALS']
                if not os.path.exists(globalConfigPath):
                    globalConfigPath = None
                    log.warning(f'PYBYTHEC_GLOBALS points to {globalConfigPath}, which doesn\'t exist')
            elif os.path.exists('.pybythecGlobals.json'):
                globalConfigPath = '.pybythecGlobals.json'
            elif os.path.exists('pybythecGlobals.json'):
                globalConfigPath = 'pybythecGlobals.json'
            else:  # check the home directory
                homeDirPath = os.path.expanduser('~')
                if os.path.exists(homeDirPath + '/.pybythecGlobals.json'):
                    globalConfigPath = homeDirPath + '/.pybythecGlobals.json'
                elif os.path.exists(homeDirPath + '/pybythecGlobals.json'):
                    globalConfigPath = homeDirPath + '/pybythecGlobals.json'
                else:  # end of the line
                    log.warning('no pybythecGlobals.json found in the home directory (hidden or otherwise)')
            if globalConfigPath and os.path.exists(globalConfigPath):

                with open(globalConfigPath, 'r') as rf:
                    self.globalConfig = json.load(rf)
                self.latestConfigTimestamp = float(os.stat(globalConfigPath).st_mtime)

                # determine default binary format
                if 'binaryFormat' in self.globalConfig:
                    self.binaryFormatDefault = self.globalConfig['binaryFormat']

        # project config
        if not self.projConfig:
            projConfigPath = ''
            if 'PYBYTHEC_PROJECT' in os.environ:
                projConfigPath = os.environ['PYBYTHEC_PROJECT']
                if not os.path.exists(projConfigPath):
                    log.warning(f'PYBYTHEC_PROJECT points to {projConfigPath}, which doesn\'t exist')
            if not os.path.exists(projConfigPath):
                projConfigPath = self.shellCwDir + '/pybythecProject.json'
                if not os.path.exists(projConfigPath):
                    projConfigPath = self.shellCwDir + '/.pybythecProject.json'   
            if os.path.exists(projConfigPath):
                with open(projConfigPath, 'r') as rf:
                    self.projConfig = json.load(rf)
                projConfigTs = float(os.stat(projConfigPath).st_mtime)
                if projConfigTs > self.latestConfigTimestamp:
                    self.latestConfigTimestamp = projConfigTs

        # local config, expected to be in the current working directory
        self.localConfig = None
        localConfigPath = ''
        if 'PYBYTHEC_LOCAL' in os.environ:
            localConfigPath = os.environ['PYBYTHEC_LOCAL']
            log.warning(f'PYBYTHEC_LOCAL points to {localConfigPath}, which doesn\'t exist')
        if not os.path.exists(localConfigPath):
            localConfigPath = self.shellCwDir + '/pybythec.json'
            if not os.path.exists(localConfigPath):
                localConfigPath = self.shellCwDir + '/.pybythec.json'
        if os.path.exists(localConfigPath):
            localConfigTs = float(os.stat(localConfigPath).st_mtime)
            if localConfigTs > self.latestConfigTimestamp:
                self.latestConfigTimestamp = localConfigTs
            with open(localConfigPath, 'r') as rf:
                    self.localConfig = json.load(rf)
        else:
            log.warning('faild to find a local pybythec.json')

        #
        # first iteration to get osType and custom keys (right now just for the compiler)
        #
        if self.globalConfig is not None:
            self._getBuildElements1(self.globalConfig)
        else:
            log.warning('not using a global configuration')
        if self.projConfig is not None:
            self._getBuildElements1(self.projConfig)
        if self.localConfig is not None:
            self._getBuildElements1(self.localConfig)
        else:
            log.warning('not using a local pybythec configuration')

        self.targetFilename = self.targetName

        if self.osTypeOverride:
            self.osType = self.osTypeOverride

        if self.osType and self.osType not in ['linux', 'macOs', 'windows']:  # validate
            log.warning(f'{self.osType} invalid osType, defaulting to the native os')
            self.osType = None

        shellOsType = utils.getShellOsType()

        if not self.osType:  # use the native os
            self.osType = shellOsType

        if self.osType == 'windows' and (shellOsType == 'linux' or shellOsType == 'macOs'):
            self.cwDir = utils.linuxToWindows(self.shellCwDir)
        else:
            self.cwDir = self.shellCwDir

    def configBuild(self, currentBuild = None):
        '''
    '''
        if currentBuild:
            self.currentBuild = currentBuild
            os.environ['PYBYTHEC_BUILDVAR'] = self.currentBuild

        # set by the config files first
        self.compiler = None  # g++-4.4 g++ clang++ msvc-11.0 etc
        self.filetype = None  # elf, mach-o, pe
        self.installDirPath = None

        self.sources = []
        self.libs = []
        # self.defines = []
        self.flags = []
        self.linkFlags = []

        self.incPaths = []
        self.extIncPaths = []  # these will not be checked for timestamps
        self.libPaths = []
        self.libSrcPaths = []

        self.qtClasses = []

        # 2 keys at this point for a potentially nested compiler: osType and currentBuild
        keys = [self.osType]
        if self.currentBuild:
            keys.append(self.currentBuild)

        #
        # second iteration to get the other configs that can't be nested (the compiler being the exception)
        #
        if self.globalConfig is not None:
            self._getBuildElements2(self.globalConfig, keys)
        if self.projConfig is not None:
            self._getBuildElements2(self.projConfig, keys)
        if self.localConfig is not None:
            self._getBuildElements2(self.localConfig, keys)

        # compiler stuff
        if self.compilerOverride:
            self.compiler = self.compilerOverride

        if not self.compiler:
            raise PybythecError('compiler not found')

        # validate compiler and determine root: can be gcc, clang or msvc
        self.compilerRoot = None
        if self.compiler.startswith('gcc') or self.compiler.startswith('g++'):
            self.compilerRoot = 'gcc'
        elif self.compiler.startswith('clang') or self.compiler.startswith('clang++'):
            self.compilerRoot = 'clang'
        elif self.compiler.startswith('msvc'):
            self.compilerRoot = 'msvc'
            if self.compiler == 'msvc':  # needs a version ie msvc-11.0 for pathing
                if self.msvcDefault:
                    self.compiler = self.msvcDefault
                else:
                    raise PybythecError('msvc has no default set, try setting the compiler to a specific version ie msvc-14.0')
        else:
            raise PybythecError(f'unrecognized compiler {self.compiler}, using the default based on osType')

        if self.buildTypeOverride:
            self.buildType = self.buildTypeOverride

        if self.binaryFormatOverride:
            self.binaryFormat = self.binaryFormatOverride

        keys += ['all', self.compilerRoot, self.compiler, self.binaryType, self.buildType, self.binaryFormat]

        if self.globalConfig is not None:
            self._getBuildElements3(self.globalConfig, keys)
        if self.projConfig is not None:
            self._getBuildElements3(self.projConfig, keys)
        if self.localConfig is not None:
            self._getBuildElements3(self.localConfig, keys)

        log.debug('PATH')
        for p in os.environ['PATH'].split(os.pathsep):
            log.debug(p)

        # deal breakers (that don't appear in the default pybythecGlobals.json)
        if not self.targetName:
            raise PybythecError('no target specified')
        if not self.binaryType:
            raise PybythecError('no binary type specified')
        if self.binaryType not in ('exe', 'static', 'dynamic', 'plugin'):
            raise PybythecError('unrecognized binary type: ' + self.binaryType)
        if not self.sources:
            raise PybythecError('no source files specified')

        #
        # compiler config
        #
        self.compilerCmd = None
        self.linker = None
        self.targetFlag = None
        self.libFlag = None
        self.libPathFlag = None
        self.objExt = None
        self.objPathFlag = None
        self.staticExt = None
        self.dynamicExt = None
        self.pluginExt = None
        self.compilerVersion = None

        # log.debug(f'{self.compilerRoot}, {self.compiler}')

        #
        # gcc / clang
        #
        if self.compilerRoot == 'gcc' or self.compilerRoot == 'clang':
            if self.plusplus:
                if self.compiler.startswith('gcc'):
                    self.compiler = self.compiler.replace('gcc', 'g++')
                elif self.compiler.startswith('clang-') or self.compiler == ('clang'):
                    self.compiler = self.compiler.replace('clang', 'clang++')

            self.compilerCmd = self.compiler
            self.objFlag = '-c'
            self.objExt = '.o'
            self.objPathFlag = '-o'
            self.defines.append('_' + self.binaryFormat.upper())

            # link
            self.linker = self.compilerCmd  # 'ld'
            self.targetFlag = '-o'
            self.libFlag = '-l'
            self.libPathFlag = '-L'
            self.staticExt = '.a'
            self.dynamicExt = '.so'
            self.pluginExt = '.so'

            if self.filetype == 'mach-o':
                self.dynamicExt = '.dylib'
                self.pluginExt = '.bundle'

            if self.binaryType == 'static' or self.binaryType == 'dynamic':
                self.targetFilename = 'lib' + self.targetName

            if self.binaryType == 'exe':
                pass
            elif self.binaryType == 'static':
                self.targetFilename = self.targetFilename + '.a'
                self.linker = 'ar'
                self.targetFlag = 'r'
            elif self.binaryType == 'dynamic':
                self.targetFilename = self.targetFilename + self.dynamicExt
            elif self.binaryType == 'plugin':
                self.targetFilename = self.targetFilename + self.pluginExt

            # get the compiler version
            log.d(self.compiler)

            compInfo = self.compiler.split('-')
            if len(compInfo) > 1:
                self.compilerVersion = self.compiler
            else:
                output = utils.runCmd([self.compiler, '-v'])
                start = output.find('gcc version') + 12
                numDots = 0
                end = start
                numDots = 0
                while True:
                    end += 1
                    if output[end] == '.':
                        numDots += 1
                    if numDots >= 2:
                        break
                v = output[start:end].strip(' ')
                self.compilerVersion = f'{self.compiler}-{v}'
            log.d(self.compilerVersion)

        #
        # msvc / msvc
        #
        elif self.compilerRoot == 'msvc':

            # compile
            self.compilerCmd = 'cl.exe'
            self.objFlag = '/c'
            self.objExt = '.obj'
            self.objPathFlag = '/Fo'

            # link
            self.linker = 'link.exe'
            self.targetFlag = '/OUT:'
            self.libFlag = ''
            self.libPathFlag = '/LIBPATH:'
            self.staticExt = '.lib'
            self.dynamicExt = '.dll'
            if self.binaryFormat == '64bit':
                self.linkFlags.append('/MACHINE:X64')

            if self.binaryType == 'exe':
                self.targetFilename = self.targetName + '.exe'
            elif self.binaryType == 'static':
                self.targetFilename = self.targetName + self.staticExt
                self.linker = 'lib.exe'
            elif self.binaryType == 'dynamic' or self.binaryType == 'plugin':
                self.targetFilename = self.targetName + self.dynamicExt
                self.linkFlags.append('/DLL')

            self.compilerVersion = self.compiler

        else:
            raise PybythecError(f'unrecognized compiler root: {self.compilerRoot}')

        # make sure the compiler is in PATH
        try:
            subprocess.call(self.compilerCmd, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        except OSError as e:
            log.error(str(e))
            for p in os.environ['PATH'].split(os.pathsep):
                log.i(p)
            raise PybythecError(f'compiler {self.compilerCmd} is not found in PATH')
            

        # make sure the linker is in PATH
        try:
            subprocess.call(self.linker, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        except OSError as e:
            log.debug(str(e))
            raise PybythecError(f'linker {self.linker} is not found in PATH')

        #
        # determine paths
        #
        installDirPath = self.installDirPath
        self.installDirPath = utils.getAbsPath(self.cwDir, installDirPath)
        self.installDirPathShell = utils.getShellPath(self.installDirPath)

        utils.resolvePaths(self.cwDir, self.sources, self.osType)
        utils.resolvePaths(self.cwDir, self.incPaths, self.osType)
        utils.resolvePaths(self.cwDir, self.extIncPaths, self.osType)
        utils.resolvePaths(self.cwDir, self.libPaths, self.osType)
        utils.resolvePaths(self.cwDir, self.libSrcPaths, self.osType)

        self.binaryRelPath = f'/{self.osType}/{self.compilerVersion}/{self.binaryFormat}/{self.buildType}'

        if self.currentBuild:
            self.binaryRelPath += '/' + self.currentBuild

        # log.debug(f'binaryRelPath: {self.binaryRelPath}')
        # log.debug(f'buildDir: {self.buildDir}')

        self.buildDirPath = utils.getAbsPath(self.cwDir, self.buildDir + self.binaryRelPath)
        self.buildDirPathShell = utils.getAbsPath(self.shellCwDir, self.buildDir + self.binaryRelPath)

        if self.libInstallPathAppend and (self.binaryType in ['static', 'dynamic']):
            self.installDirPath += self.binaryRelPath
            self.installDirPathShell += self.binaryRelPath

        self.installPath = f'{self.installDirPath}/{self.targetFilename}'
        self.installPathShell = f'{self.installDirPathShell}/{self.targetFilename}'

        log.debug(f'buildDirPath: {self.buildDirPath}')
        log.debug(f'buildDirPathShell: {self.buildDirPathShell}')
        log.debug(f'installDirPath: {self.installDirPath}')
        log.debug(f'installDirPathShell: {self.installDirPathShell}')
        log.debug(f'installPath: {self.installPath}')
        log.debug(f'installPathShell: {self.installPathShell}')
        # raise PybythecError('early return')

        self.infoStr = f'{self.targetName} ({self.osType} {self.compilerVersion} {self.binaryFormat} {self.buildType}'
        if self.currentBuild:
            self.infoStr += f' {self.currentBuild}'
        self.infoStr += ')'

    def _getBuildElements1(self, configObj):
        '''
      elements that aren't nested
    '''
        if 'buildType' in configObj:
            self.buildType = os.path.expandvars(configObj['buildType'])

        if 'version' in configObj:
            self.version = os.path.expandvars(configObj['version'])

        if 'target' in configObj:
            self.targetName = os.path.expandvars(configObj['target'])

        if 'builds' in configObj:
            self.builds = configObj['builds']

        if 'osType' in configObj:
            self.osType = os.path.expandvars(configObj['osType'])

        if 'binaryType' in configObj:
            self.binaryType = os.path.expandvars(configObj['binaryType'])

        if 'binaryFormat' in configObj:
            self.binaryFormat = os.path.expandvars(configObj['binaryFormat'])

        if 'libInstallPathAppend' in configObj:
            self.libInstallPathAppend = configObj['libInstallPathAppend']

        if 'plusplus' in configObj:
            self.plusplus = configObj['plusplus']

        if 'locked' in configObj:
            self.locked = configObj['locked']

        if 'buildDir' in configObj:
            self.buildDir = configObj['buildDir']

        if 'showCompilerCmds' in configObj:
            self.showCompilerCmds = configObj['showCompilerCmds']

        if 'showLinkerCmds' in configObj:
            self.showLinkerCmds = configObj['showLinkerCmds']

        if 'copyDynamicLibs' in configObj:
            self.copyDynamicLibs = configObj['copyDynamicLibs']

        if 'msvc-default' in configObj:
            self.msvcDefault = configObj['msvc-default']

    def _getBuildElements2(self, configObj, keys = []):
        '''
        elements that are nested in a finite / special case way, currently just the compiler
        '''
        # compiler can be nested in a dict with 2 valid key types: osType and a build name
        if 'compiler' in configObj:
            compilerList = []
            self._getArgsList(compilerList, configObj['compiler'], keys)
            if len(compilerList):
                self.compiler = compilerList[0]
                if len(compilerList) > 1:
                    log.warning(f'couldn\'t resolve to single compiler, compiler options: {compilerList}, selecting {self.compiler}')

    def _getBuildElements3(self, configObj, keys = []):
        '''
            elements that are potentially nested in any which way
        '''
        if 'bins' in configObj:
            binPaths = []
            self._getArgsList(binPaths, configObj['bins'], keys)
            for binPath in binPaths:
                if not utils.pathExists(binPath):
                    continue
                binPath = utils.getShellPath(binPath)
                os.environ['PATH'] = binPath + os.pathsep + os.environ['PATH']

        if 'sources' in configObj:
            self._getArgsList(self.sources, configObj['sources'], keys)

        if 'libs' in configObj:
            self._getArgsList(self.libs, configObj['libs'], keys)

        if 'defines' in configObj:
            self._getArgsList(self.defines, configObj['defines'], keys)

        if 'flags' in configObj:
            self._getArgsList(self.flags, configObj['flags'], keys)

        if 'linkFlags' in configObj:
            self._getArgsList(self.linkFlags, configObj['linkFlags'], keys)

        if 'incPaths' in configObj:
            self._getArgsList(self.incPaths, configObj['incPaths'], keys)

        if 'extIncPaths' in configObj:
            self._getArgsList(self.extIncPaths, configObj['extIncPaths'], keys)

        if 'libPaths' in configObj:
            self._getArgsList(self.libPaths, configObj['libPaths'], keys)

        if 'libSrcPaths' in configObj:
            self._getArgsList(self.libSrcPaths, configObj['libSrcPaths'], keys)

        if 'qtClasses' in configObj:
            self._getArgsList(self.qtClasses, configObj['qtClasses'], keys)

        if 'filetype' in configObj:
            filetypes = []
            self._getArgsList(filetypes, configObj['filetype'], keys)
            if len(filetypes):
                self.filetype = filetypes[0]

        if 'installPath' in configObj:
            installPaths = []
            self._getArgsList(installPaths, configObj['installPath'], keys)
            if len(installPaths):
                self.installDirPath = installPaths[0]
        
        if 'processType' in configObj:
            processTypes = []
            self._getArgsList(processTypes, configObj['processType'], keys)
            if len(processTypes):
                processType = processTypes[0]
                if processType == 'multi-process':
                    self.processType = utils.MULTI_PROCESS
                elif processType == 'multi-thread':
                    self.processType = utils.MULTI_THREAD
                elif processType == 'linear-process':
                    self.processType = utils.LINEAR_PROCESS


    def _getArgsList(self, argsList, args, keys = []):
        '''
      recursivley parses args and appends it to argsList if it has any of the keys
      args can be a dict, str (space-deliminated) or list
    '''
        if type(args) == dict:
            for key in keys:
                if key in args:
                    self._getArgsList(argsList, args[key], keys)
        else:
            if type(args) == str or type(args).__name__ == 'unicode':
                argsList.append(os.path.expandvars(args))
            elif type(args) == list:
                for arg in args:
                    argsList.insert(0, os.path.expandvars(arg))
